# Generated by Django 4.1.8 on 2023-04-12 12:56

import django.contrib.postgres.fields.ranges
import wbcore.models.fields
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="CalendarItem",
            fields=[
                (
                    "id",
                    models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
                ),
                ("is_active", models.BooleanField(default=True)),
                ("deletion_datetime", models.DateTimeField(blank=True, null=True)),
                ("color", wbcore.contrib.color.fields.ColorField(blank=True, null=True)),
                ("entity_list", models.JSONField(blank=True, null=True)),
                ("icon", models.CharField(blank=True, max_length=64, null=True)),
                (
                    "visibility",
                    models.CharField(
                        choices=[
                            ("PUBLIC", "Public"),
                            ("PRIVATE", "Private"),
                            ("CONFIDENTIAL", "Confidential"),
                        ],
                        default="PUBLIC",
                        max_length=255,
                        verbose_name="Visibility",
                    ),
                ),
                (
                    "item_type",
                    models.CharField(
                        choices=[
                            ("ABSENCE_REQUEST", "Absence Request"),
                            ("HOLIDAY", "Holiday"),
                            ("ACTIVITY", "Activity"),
                            ("TASK", "Task"),
                        ],
                        default="ACTIVITY",
                        max_length=255,
                        verbose_name="Type",
                    ),
                ),
                (
                    "period",
                    django.contrib.postgres.fields.ranges.DateTimeRangeField(
                        blank=True, null=True, verbose_name="Period"
                    ),
                ),
                ("title", models.CharField(max_length=255, verbose_name="Title")),
                (
                    "conference_room",
                    models.BooleanField(
                        default=False,
                        help_text="Check this if it takes place in the conference room",
                        verbose_name="Conference Room",
                    ),
                ),
                (
                    "all_day",
                    models.BooleanField(
                        default=False,
                        help_text="Check this if the activity spans throughout the day",
                        verbose_name="All Day",
                    ),
                ),
                ("is_cancelled", models.BooleanField(default=False, verbose_name="Is Cancelled")),
            ],
            options={
                "verbose_name": "Calendar Item",
                "verbose_name_plural": "Calendar Items",
                "permissions": [("administrate_confidential_items", "Can see confidential calendar items")],
            },
        )
    ]



from wbcore.menus import ItemPermission, MenuItem

GEOGRAPHY_MENUITEM = MenuItem(
    label="Geography",
    endpoint="geography:geography-list",
    add=MenuItem(label="New Geography", endpoint="geography:geography-list"),
    permission=ItemPermission(method=lambda request: request.user.is_staff),
)



from django.urls import include, path
from wbcore.contrib.notifications import viewsets
from wbcore.contrib.notifications.views import (
    NotificationTokenAPIView,
    service_worker_view,
)
from wbcore.routers import WBCoreRouter

router = WBCoreRouter()

router.register("notification", viewsets.NotificationModelViewSet, basename="notification")
router.register(
    "notification_type_representation",
    viewsets.NotificationTypeRepresentationViewSet,
    basename="notification_type_representation",
)
router.register(
    "notification_type_setting", viewsets.NotificationTypeSettingModelViewSet, basename="notification_type_setting"
)


urlpatterns = [
    path("service-worker/<str:service_worker_name>", service_worker_view, name="service-worker"),
    path("token/", NotificationTokenAPIView.as_view(), name="token"),
    path("", include(router.urls)),
]



from wbcore.metadata.configs.display.instance_display.layouts.inlines import Inline


class TestInline:
    def test_init(self):
        inline = Inline(key="key", endpoint="endpoint")
        assert inline is not None

    def test_serialize(self):
        inline = Inline(key="key", endpoint="endpoint")
        assert inline.serialize() == {"key": "key", "endpoint": "endpoint", "hide_controls": False}

    # TODO: Test with display



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class UserActivityModelEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None

    def get_list_endpoint(self, **kwargs):
        return reverse("wbcore:authentication:useractivity-list", args=[], request=self.request)


class UserActivityUserModelEndpointConfig(UserActivityModelEndpointConfig):
    def get_list_endpoint(self, **kwargs):
        return reverse(
            "wbcore:authentication:user-useractivity-list", args=[self.view.kwargs["user_id"]], request=self.request
        )


class UserActivityTableEndpointConfig(UserActivityModelEndpointConfig):
    def get_list_endpoint(self, **kwargs):
        return reverse("wbcore:authentication:useractivitytable-list", args=[], request=self.request)



from django.urls import include, path
from wbcore.routers import WBCoreRouter

from . import viewsets

router = WBCoreRouter()

router.register(r"league", viewsets.LeagueModelViewSet, basename="league")
router.register(r"player", viewsets.PlayerModelViewSet, basename="player")
router.register(r"treeviewplayer", viewsets.TreeViewPlayerModelViewSet, basename="treeviewplayer")
router.register(r"person", viewsets.SportPersonModelViewSet, basename="person")
router.register(r"persontooltip", viewsets.SportPersonToolTipViewset, basename="persontooltip")
router.register(r"stadium", viewsets.StadiumModelViewSet, basename="stadium")
router.register(r"season", viewsets.SeasonModelViewSet, basename="season")
router.register(r"team", viewsets.TeamModelViewSet, basename="team")
router.register(r"sport", viewsets.SportModelViewSet, basename="sport")
router.register(r"role", viewsets.RoleModelViewSet, basename="role")
router.register(r"match", viewsets.MatchModelViewSet, basename="match")
router.register(r"event", viewsets.EventModelViewSet, basename="event")
router.register(r"eventtype", viewsets.EventTypeModelViewSet, basename="eventtype")
router.register(r"teamresults", viewsets.TeamResultsModelViewSet, basename="teamresults")

router.register(r"personrepresentation", viewsets.SportPersonRepresentationViewSet, basename="personrepresentation")
router.register(r"stadiumrepresentation", viewsets.StadiumRepresentationViewSet, basename="stadiumrepresentation")
router.register(r"seasonrepresentation", viewsets.SeasonRepresentationViewSet, basename="seasonrepresentation")
router.register(r"leaguerepresentation", viewsets.LeagueRepresentationViewSet, basename="leaguerepresentation")
router.register(r"sportrepresentation", viewsets.SportRepresentationViewSet, basename="sportrepresentation")
router.register(r"teamrepresentation", viewsets.TeamRepresentationViewSet, basename="teamrepresentation")
router.register(r"rolerepresentation", viewsets.RoleRepresentationViewSet, basename="rolerepresentation")
router.register(r"playerrepresentation", viewsets.PlayerRepresentationViewSet, basename="playerrepresentation")
router.register(r"matchrepresentation", viewsets.MatchRepresentationViewSet, basename="matchrepresentation")
router.register(r"eventrepresentation", viewsets.EventRepresentationViewSet, basename="eventrepresentation")
router.register(
    r"eventtyperepresentation", viewsets.EventTypeRepresentationViewSet, basename="eventtyperepresentation"
)
router.register(
    r"teamresultsrepresentation", viewsets.TeamResultsRepresentationViewSet, basename="teamresultsrepresentation"
)

stadium_router = WBCoreRouter()
stadium_router.register(r"team-stadium", viewsets.TeamStadiumModelViewSet, basename="team-stadium")
stadium_router.register(r"match-stadium", viewsets.MatchStadiumModelViewSet, basename="match-stadium")

sport_router = WBCoreRouter()
sport_router.register(r"league-sport", viewsets.LeagueSportModelViewSet, basename="league-sport")
sport_router.register(r"eventtype-sport", viewsets.EventTypeSportModelViewSet, basename="eventtype-sport")

team_router = WBCoreRouter()
team_router.register(r"player-team", viewsets.PlayerTeamModelViewSet, basename="player-team")
team_router.register(r"match-team", viewsets.MatchModelViewSet, basename="match-team")

match_router = WBCoreRouter()
match_router.register(r"event-match", viewsets.EventMatchModelViewSet, basename="event-match")

league_router = WBCoreRouter()
league_router.register(r"match-league", viewsets.MatchLeagueModelViewSet, basename="match-league")
league_router.register(r"teamresults-league", viewsets.TeamResultsLeagueModelViewSet, basename="teamresults-league")

league_event_type_router = WBCoreRouter()
league_event_type_router.register(
    r"league-player-statistics", viewsets.LeaguePlayerStatisticsModelViewSet, basename="league-player-statistics"
)
league_event_type_router.register(
    r"league-team-statistics", viewsets.LeagueTeamStatisticsModelViewSet, basename="league-team-statistics"
)

player_router = WBCoreRouter()
player_router.register(r"player-statistics", viewsets.PlayerStatisticsChartModelViewSet, basename="player-statistics")

urlpatterns = [
    path("", include(router.urls)),
    path("stadium/<int:stadium_id>/", include(stadium_router.urls)),
    path("sport/<int:sport_id>/", include(sport_router.urls)),
    path("team/<int:team_id>/", include(team_router.urls)),
    path("match/<int:match_id>/", include(match_router.urls)),
    path("league/<int:league_id>/", include(league_router.urls)),
    path("league/<int:league_id>/event_type/<int:event_type_id>/", include(league_event_type_router.urls)),
    path("player/<int:player_id>/", include(player_router.urls)),
]



from django.utils.translation import gettext_lazy as _
from wbcore import filters
from wbcore.contrib.agenda.models import Building, ConferenceRoom
from wbcore.contrib.directory.models import AddressContact


class ConferenceRoomFilter(filters.FilterSet):
    building = filters.ModelMultipleChoiceFilter(
        label=_("Buildings"),
        queryset=Building.objects.all(),
        endpoint=Building.get_representation_endpoint(),
        value_key=Building.get_representation_value_key(),
        label_key=Building.get_representation_label_key(),
        field_name="building",
    )

    class Meta:
        model = ConferenceRoom
        fields = {
            "name": ["exact", "icontains"],
            "is_videoconference_capable": ["exact"],
            "email": ["exact", "icontains"],
            "capacity": ["exact", "lte", "gte"],
        }


class BuildingFilter(filters.FilterSet):
    address = filters.ModelMultipleChoiceFilter(
        label=_("Addresses"),
        queryset=AddressContact.objects.all(),
        endpoint=AddressContact.get_representation_endpoint(),
        value_key=AddressContact.get_representation_value_key(),
        label_key=AddressContact.get_representation_label_key(),
        field_name="address",
    )

    class Meta:
        model = Building
        fields = {
            "name": ["exact", "icontains"],
        }



from .components import Component


class Offset(Component):
    name_part: str

    def __init__(self, name_part: str, amount: int = 1):
        self.name_part = name_part
        self.amount = amount

    def __call__(self, amount: int) -> "Offset":
        return self.__class__(self.name_part, amount)

    @property
    def name(self) -> str:
        return f"{self.amount}{self.name_part}"


Second = Offset("seconds")
Minute = Offset("minutes")
Hour = Offset("hours")
Day = Offset("days")
BusinessDay = Offset("bdays")
Week = Offset("weeks")
Month = Offset("months")
Quarter = Offset("quarters")
Year = Offset("years")



from django.conf import settings
from django.utils.module_loading import import_string
from rest_framework.request import Request
from rest_framework.reverse import reverse
from wbcore.configs.decorators import register_config
from wbcore.shares.config import share_action_button


@register_config
def release_note_config(request: Request) -> tuple[str, dict[str, str]]:
    return "release_notes", {
        "endpoint": reverse("wbcore:releasenote-list", request=request),
        "unread_release_notes": reverse("wbcore:releasenote-unread-count", request=request),
    }


@register_config
def menu_config(request: Request) -> tuple[str, str]:
    return "menu", reverse("wbcore:menu", request=request)


@register_config
def share_config(request: Request):
    return "share", share_action_button(request=request).serialize(request)


@register_config
def markdown_config(request: Request) -> tuple[str, dict[str, str]]:
    return "markdown", {"blockdiag": reverse("wbcore:blockdiag", request=request)}


@register_config
def menu_calendar_config(request: Request) -> tuple[str, str | None]:
    menu_calendar = None
    if settings.FRONTEND_MENU_CALENDAR:
        menu_calendar = import_string(settings.FRONTEND_MENU_CALENDAR)(request=request)
    return "menu_calendar", menu_calendar


@register_config
def beta_button(request: Request) -> tuple[str, dict] | None:
    beta_url = settings.BETA_BUTTON_URL
    if ((beta_version := settings.BETA_BUTTON_VERSION) or beta_url) and (beta_text := settings.BETA_BUTTON_TEXT):
        if beta_version and not beta_url:
            beta_url = f"{settings.CDN_BASE_ENDPOINT_URL}/{beta_version}/main.js"
        return "beta_button", {"url": beta_url, "text": beta_text}


@register_config
def frontend_version(request: Request) -> tuple[str, str | None]:
    return "frontend_version", getattr(settings, "FRONTEND_VERSION", None)



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class TransitionEndpointConfig(EndpointViewConfig):
    def get_create_endpoint(self, **kwargs) -> str:
        if step_id := self.view.kwargs.get("step_id"):
            return reverse("wbcore:workflow:transition-step-list", args=[step_id], request=self.request)
        if workflow_id := self.view.kwargs.get("workflow_id"):
            return reverse("wbcore:workflow:transition-workflow-list", args=[workflow_id], request=self.request)
        return super().get_create_endpoint(**kwargs)

    def get_instance_endpoint(self, **kwargs):
        if step_id := self.view.kwargs.get("step_id"):
            return reverse("wbcore:workflow:transition-step-list", args=[step_id], request=self.request)
        if workflow_id := self.view.kwargs.get("workflow_id"):
            return reverse("wbcore:workflow:transition-workflow-list", args=[workflow_id], request=self.request)
        return super().get_instance_endpoint(**kwargs)



import pytest
from rest_framework.fields import Field
from wbcore.serializers.fields.mixins import WBCoreSerializerFieldMixin


class TestListField:
    def setup_method(self):
        class TestSubClass(WBCoreSerializerFieldMixin, Field):
            field_type = "field_type"

            def to_representation(self, value):
                return value

        self.TestSubClass = TestSubClass

    def test_init(self):
        test_sub_class_instance = self.TestSubClass()
        assert test_sub_class_instance is not None

    def test_init_with_extra(self):
        test_sub_class_instance = self.TestSubClass(extra={})
        assert test_sub_class_instance is not None

    def test_init_with_decorators(self):
        test_sub_class_instance = self.TestSubClass(decorators={})
        assert test_sub_class_instance is not None

    @pytest.mark.parametrize("label", ["", "Label", None])
    @pytest.mark.parametrize("required", [True, False])
    @pytest.mark.parametrize("read_only", [False])
    def test_get_representation(self, label, required, read_only):
        test_sub_class_instance = self.TestSubClass(
            label=label,
            required=required,
            read_only=read_only,
        )
        assert test_sub_class_instance.get_representation(None, "field_name") == (
            "field_name",
            {
                "key": "field_name",
                "label": label,
                "type": "field_type",
                "required": required,
                "read_only": read_only,
                "decorators": [],
                "depends_on": [],
            },
        )

    def test_get_representation_additional_attrs(self):
        help_text = "This is a help text"
        decorators = {"some_field": [{"position": "right", "value": "%"}]}
        extra = {"some_key": "some_value"}

        test_sub_class_instance = self.TestSubClass(
            help_text=help_text,
            decorators=decorators,
            extra=extra,
        )
        assert test_sub_class_instance.get_representation(None, "field_name") == (
            "field_name",
            {
                "key": "field_name",
                "label": None,
                "type": "field_type",
                "required": True,
                "read_only": False,
                "help_text": help_text,
                "decorators": decorators,
                "extra": extra,
                "depends_on": [],
            },
        )

    def test_get_representation_additional_default_value(self):
        default_value = 100
        test_sub_class_instance = self.TestSubClass(default=default_value)

        assert test_sub_class_instance.get_representation(None, "field_name") == (
            "field_name",
            {
                "key": "field_name",
                "label": None,
                "type": "field_type",
                "required": False,  # If a default is given, then it cannot be required
                "read_only": False,
                "default": default_value,
                "decorators": [],
                "depends_on": [],
            },
        )

    def test_get_representation_additional_default_method(self):
        def default_method():
            return 100

        test_sub_class_instance = self.TestSubClass(default=default_method)

        assert test_sub_class_instance.get_representation(None, "field_name") == (
            "field_name",
            {
                "key": "field_name",
                "label": None,
                "type": "field_type",
                "required": False,  # If a default is given, then it cannot be required
                "read_only": False,
                "default": default_method(),
                "decorators": [],
                "depends_on": [],
            },
        )



import json

import pytest
from rest_framework import status
from rest_framework.test import APIRequestFactory
from wbcore.contrib.authentication.factories import UserFactory
from wbcore.contrib.directory.models import ClientManagerRelationship as CMR
from wbcore.contrib.directory.models import Person
from wbcore.test.utils import (
    get_data_from_factory,
    get_kwargs,
    get_model_factory,
    get_or_create_superuser,
)

from ..factories import ClientManagerRelationshipFactory, EntryFactory, PersonFactory
from ..viewsets import (
    AddressContactEntryViewSet,
    BankingContactEntryViewSet,
    ClientManagerViewSet,
    CompanyModelViewSet,
    CompanyRepresentationViewSet,
    CompanyTypeModelViewSet,
    CustomerStatusModelViewSet,
    EmailContactEntryViewSet,
    EntryRepresentationViewSet,
    PersonRepresentationViewSet,
    PositionModelViewSet,
    RelationshipModelViewSet,
    RelationshipTypeModelViewSet,
    SocialMediaContactEntryViewSet,
    SpecializationModelViewSet,
    TelephoneContactEntryViewSet,
    UserIsClientViewSet,
    UserIsManagerViewSet,
    WebsiteContactEntryViewSet,
)


@pytest.mark.django_db
class TestEntryModelViewSet:
    api_factory = APIRequestFactory()

    @pytest.mark.parametrize(
        "rmvs",
        [
            EntryRepresentationViewSet,
            PersonRepresentationViewSet,
            CompanyRepresentationViewSet,
        ],
    )
    def test_get_queryset(self, rmvs):
        request = APIRequestFactory().get("")
        request.user = UserFactory(is_active=True, is_superuser=False)
        factory = get_model_factory(rmvs.queryset.model)
        obj = factory()
        kwargs = get_kwargs(obj, rmvs, request)
        vs = rmvs.as_view({"get": "list"})
        response = vs(request, **kwargs)
        assert response.status_code == status.HTTP_403_FORBIDDEN

    @pytest.mark.parametrize("rmvs", [CompanyRepresentationViewSet])
    def test_get_queryset_CompanyRepresentationViewSet(self, rmvs, company_factory):
        request = APIRequestFactory().get("")
        request.user = UserFactory(is_active=True, is_superuser=False)
        request.user.profile.employers.add(company_factory())
        factory = get_model_factory(rmvs.queryset.model)
        obj = factory()
        kwargs = get_kwargs(obj, rmvs, request)
        vs = rmvs.as_view({"get": "list"})
        response = vs(request, **kwargs)
        assert response.status_code == status.HTTP_403_FORBIDDEN

    @pytest.mark.parametrize("mvs", [CompanyModelViewSet])
    def test_get_company_instance(self, mvs, company_factory):
        # Arrange
        company = company_factory()
        request = self.api_factory.get("")
        request.user = UserFactory(is_active=True, is_superuser=True)
        view = mvs.as_view({"get": "retrieve"})
        # Act
        response = view(request, pk=company.id).render()
        # Assert
        assert response.status_code == status.HTTP_200_OK
        assert response.data.get("instance")
        assert not response.data.get("results")
        assert response.data["instance"]["id"] == company.id

    @pytest.mark.parametrize("mvs", [CompanyModelViewSet])
    def test_get_company_list(self, mvs, company_factory):
        # Arrange
        company_factory.create_batch(3)
        request = self.api_factory.get("")
        request.user = UserFactory(is_active=True, is_superuser=True)
        view = mvs.as_view({"get": "list"})
        # Act
        response = view(request).render()
        # Assert
        assert response.status_code == status.HTTP_200_OK
        assert not response.data.get("instance")
        assert response.data.get("results")
        assert response.data["results"]
        assert len(response.data["results"]) == 3

    @pytest.mark.parametrize("mvs", [CompanyModelViewSet])
    def test_delete_company_instance(self, mvs, company_factory):
        # Arrange
        company = company_factory()
        request = self.api_factory.delete("")
        request.user = UserFactory(is_active=True, is_superuser=True)
        view = mvs.as_view({"delete": "destroy"})
        # Act
        response = view(request, pk=company.id).render()
        # Assert
        assert response.status_code == status.HTTP_204_NO_CONTENT

    @pytest.mark.parametrize(
        "mvs",
        [
            EmailContactEntryViewSet,
            AddressContactEntryViewSet,
            TelephoneContactEntryViewSet,
            WebsiteContactEntryViewSet,
            BankingContactEntryViewSet,
        ],
    )
    def test_post_delete_contact(self, mvs):
        factory = get_model_factory(mvs.queryset.model)
        obj = factory()
        factory(entry=obj.entry)
        user = get_or_create_superuser()
        data = get_data_from_factory(obj, mvs, delete=True, superuser=user)
        request = APIRequestFactory().post("", data)
        request.user = user
        kwargs = get_kwargs(obj, mvs, request=request, data=data)
        vs = mvs.as_view({"post": "create"})
        response = vs(request, **kwargs)
        assert response.status_code == status.HTTP_201_CREATED
        assert response.data.get("instance")


# # =====================================================================================================================
# #                                                  TESTING UTILS VIEWSETS
# # =====================================================================================================================


@pytest.mark.django_db
class TestUtilsViewSets:
    api_factory = APIRequestFactory()

    @pytest.mark.parametrize(
        "mvs",
        [
            CustomerStatusModelViewSet,
            PositionModelViewSet,
            CompanyTypeModelViewSet,
            SpecializationModelViewSet,
        ],
    )
    def test_get_utils(self, mvs):
        request = self.api_factory.get("")
        request.user = UserFactory(is_active=True, is_superuser=True)
        factory = get_model_factory(mvs.queryset.model)
        factory.create_batch(3)
        vs = mvs.as_view({"get": "list"})
        response = vs(request)
        assert response.data.get("results")
        assert not response.data.get("instance")
        assert len(response.data.get("results")) == 3
        assert response.status_code == status.HTTP_200_OK

    @pytest.mark.parametrize(
        "mvs",
        [
            CustomerStatusModelViewSet,
            PositionModelViewSet,
            CompanyTypeModelViewSet,
            SpecializationModelViewSet,
        ],
    )
    def test_retrieve_utils(self, mvs):
        request = self.api_factory.get("")
        request.user = UserFactory(is_active=True, is_superuser=True)
        factory = get_model_factory(mvs.queryset.model)
        obj = factory()
        vs = mvs.as_view({"get": "retrieve"})
        response = vs(request, pk=obj.id)
        assert response.data.get("instance")
        assert not response.data.get("results")
        assert response.status_code == status.HTTP_200_OK

    @pytest.mark.parametrize(
        "mvs",
        [
            CustomerStatusModelViewSet,
            PositionModelViewSet,
            CompanyTypeModelViewSet,
            SpecializationModelViewSet,
        ],
    )
    def test_post_utils(self, mvs):
        factory = get_model_factory(mvs.queryset.model)
        obj = factory()
        super_user = UserFactory(is_active=True, is_superuser=True)
        data = get_data_from_factory(obj, mvs, superuser=super_user, delete=True)
        request = self.api_factory.post("", data=data)
        request.user = super_user
        kwargs = get_kwargs(obj, mvs, request)
        vs = mvs.as_view({"post": "create"})
        response = vs(request, **kwargs)
        assert response.status_code == status.HTTP_201_CREATED

    @pytest.mark.parametrize(
        "mvs",
        [
            CustomerStatusModelViewSet,
            PositionModelViewSet,
            CompanyTypeModelViewSet,
            SpecializationModelViewSet,
        ],
    )
    def test_delete_utils(self, mvs):
        request = self.api_factory.delete("")
        request.user = UserFactory(is_active=True, is_superuser=True)
        factory = get_model_factory(mvs.queryset.model)
        obj = factory()
        kwargs = get_kwargs(obj, mvs, request)
        vs = mvs.as_view({"delete": "destroy"})
        response = vs(request, **kwargs, pk=obj.pk)
        assert response.status_code == status.HTTP_204_NO_CONTENT

    @pytest.mark.parametrize(
        "mvs",
        [
            CustomerStatusModelViewSet,
            PositionModelViewSet,
            CompanyTypeModelViewSet,
            SpecializationModelViewSet,
        ],
    )
    def test_put_utils(self, mvs):
        factory = get_model_factory(mvs.queryset.model)
        old_obj = factory()
        new_obj = factory()
        user = UserFactory(is_active=True, is_superuser=True)
        data = get_data_from_factory(new_obj, mvs, superuser=user, delete=True)
        request = APIRequestFactory().put("", data=data)
        request.user = user
        vs = mvs.as_view({"put": "update"})
        response = vs(request, pk=old_obj.id)
        assert response.status_code == status.HTTP_200_OK
        assert response.data["instance"]["title"] == new_obj.title
        assert not response.data["instance"]["title"] == old_obj.title

    @pytest.mark.parametrize(
        "mvs",
        [
            CustomerStatusModelViewSet,
            PositionModelViewSet,
            CompanyTypeModelViewSet,
            SpecializationModelViewSet,
        ],
    )
    def test_patch_utils(self, mvs):
        factory = get_model_factory(mvs.queryset.model)
        obj = factory()
        request = APIRequestFactory().patch("", data={"title": "New Title"})
        request.user = UserFactory(is_active=True, is_superuser=True)
        vs = mvs.as_view({"patch": "partial_update"})
        response = vs(request, pk=obj.id)
        assert response.status_code == status.HTTP_200_OK
        assert response.data["instance"]["title"] == "New Title"


# =====================================================================================================================
#                                                  TESTING RELATIONSHIP VIEWSETS
# =====================================================================================================================


@pytest.mark.django_db
class TestRelationshipViewSets:
    api_factory = APIRequestFactory()

    @pytest.mark.parametrize("mvs", [RelationshipTypeModelViewSet])
    def test_relationshiptype_retrieve(self, mvs):
        request = APIRequestFactory().get("")
        request.user = UserFactory(is_active=True, is_superuser=True)
        factory = get_model_factory(mvs.queryset.model)
        type = factory()
        vs = mvs.as_view({"get": "retrieve"})
        response = vs(request, pk=type.id)
        assert response.data.get("instance")
        assert not response.data.get("results")
        assert response.status_code == status.HTTP_200_OK

    @pytest.mark.parametrize("mvs", [RelationshipTypeModelViewSet])
    def test_relationshiptype_get(self, mvs):
        request = APIRequestFactory().get("")
        request.user = UserFactory(is_active=True, is_superuser=True)
        factory = get_model_factory(mvs.queryset.model)
        factory.create_batch(3)
        vs = mvs.as_view({"get": "list"})
        response = vs(request)
        assert response.data.get("results")
        assert not response.data.get("instance")
        assert len(response.data.get("results")) == 6
        assert response.status_code == status.HTTP_200_OK

    @pytest.mark.parametrize("mvs", [RelationshipTypeModelViewSet])
    def test_relationshiptype_put(self, mvs):
        factory = get_model_factory(mvs.queryset.model)
        old_type = factory()
        non_counter_type = factory(counter_relationship=None)
        new_type = factory()
        new_type.counter_relationship = non_counter_type
        user = UserFactory(is_active=True, is_superuser=True)
        data = get_data_from_factory(new_type, mvs, superuser=user, delete=True)
        request = APIRequestFactory().put("", data=data)
        request.user = user
        vs = mvs.as_view({"put": "update"})
        response = vs(request, pk=old_type.id)
        assert response.status_code == status.HTTP_200_OK
        assert response.data["instance"]["title"] == new_type.title
        assert not response.data["instance"]["title"] == old_type.title
        assert response.data["instance"]["counter_relationship"] == new_type.counter_relationship.id

    @pytest.mark.parametrize("mvs", [RelationshipTypeModelViewSet])
    def test_relationshiptype_delete(self, mvs):
        request = APIRequestFactory().delete("")
        request.user = UserFactory(is_active=True, is_superuser=True)
        factory = get_model_factory(mvs.queryset.model)
        type = factory()
        vs = mvs.as_view({"delete": "destroy"})
        response = vs(request, pk=type.id)
        assert response.status_code == status.HTTP_204_NO_CONTENT

    @pytest.mark.parametrize("mvs", [RelationshipTypeModelViewSet])
    def test_relationshiptype_patch(self, mvs):
        factory = get_model_factory(mvs.queryset.model)
        type = factory()
        request = APIRequestFactory().patch("", data={"title": "New Title"})
        request.user = UserFactory(is_active=True, is_superuser=True)
        vs = mvs.as_view({"patch": "partial_update"})
        response = vs(request, pk=type.id)
        assert response.status_code == status.HTTP_200_OK
        assert response.data["instance"]["title"] == "New Title"

    @pytest.mark.parametrize("mvs", [RelationshipTypeModelViewSet])
    def test_relationshiptype_post(self, mvs):
        factory = get_model_factory(mvs.queryset.model)
        type = factory(counter_relationship=None)
        user = UserFactory(is_active=True, is_superuser=True)
        data = get_data_from_factory(type, mvs, superuser=user, delete=True)
        json_data = json.dumps(data)
        request = APIRequestFactory().post("", data=json_data, content_type="application/json")
        request.user = user
        kwargs = get_kwargs(type, mvs, request)
        vs = mvs.as_view({"post": "create"})
        response = vs(request, **kwargs)
        assert response.status_code == status.HTTP_201_CREATED

    def test_relationship_list(self, relationship_factory, relationship_type_factory):
        # Arrange
        relationship_type = relationship_type_factory(counter_relationship=None)
        request = self.api_factory.get("")
        request.user = UserFactory(is_active=True, is_superuser=True)

        relationship_factory.create_batch(3, relationship_type=relationship_type)
        view = RelationshipModelViewSet.as_view({"get": "list"})
        # Act
        response = view(request).render()
        # Assert
        assert response.status_code == status.HTTP_200_OK
        assert response.data.get("results")
        assert len(response.data["results"]) == 3 * 2

    def test_relationship_retrieve(self, relationship_factory):
        # Arrange
        relationship = relationship_factory()
        request = self.api_factory.get("")
        request.user = UserFactory(is_active=True, is_superuser=True)
        view = RelationshipModelViewSet.as_view({"get": "retrieve"})
        # Act
        response = view(request, pk=relationship.id).render()
        # Assert
        assert response.status_code == status.HTTP_200_OK
        assert json.loads(response.content)["instance"]["from_entry"] == relationship.from_entry.id
        assert json.loads(response.content)["instance"]["to_entry"] == relationship.to_entry.id
        assert json.loads(response.content)["instance"]["relationship_type"] == relationship.relationship_type.id

    def test_relationship_delete(self, relationship_factory):
        # Arrange
        relationship = relationship_factory()
        request = self.api_factory.delete("")
        request.user = UserFactory(is_active=True, is_superuser=True)
        view = RelationshipModelViewSet.as_view({"delete": "destroy"})
        # Act
        response = view(request, pk=relationship.id).render()
        # Assert
        assert response.status_code == status.HTTP_204_NO_CONTENT

    def test_relationship_create(self, relationship_factory):
        # Arrange
        relationship = relationship_factory()
        user = UserFactory(is_active=True, is_superuser=True)
        data = get_data_from_factory(relationship, RelationshipModelViewSet, superuser=user)
        request = self.api_factory.post("", data=data)
        request.user = user
        kwargs = get_kwargs(relationship, RelationshipModelViewSet, request=request, data=data)
        view = RelationshipModelViewSet.as_view({"post": "create"})
        # Act
        response = view(request, kwargs).render()
        # Assert
        assert response.status_code == status.HTTP_201_CREATED

    def test_relationship_update(self, relationship_factory):
        # Arrange
        old_relationship = relationship_factory()
        new_relationship = relationship_factory()
        user = UserFactory(is_active=True, is_superuser=True)
        data = get_data_from_factory(new_relationship, RelationshipModelViewSet, superuser=user)
        request = self.api_factory.put("", data=data)
        request.user = user
        view = RelationshipModelViewSet.as_view({"put": "update"})
        # Act
        response = view(request, pk=old_relationship.id).render()
        # Assert
        assert response.status_code == status.HTTP_200_OK

    def test_relationship_partial_update(self, relationship_factory, person_factory):
        # Arrange
        relationship = relationship_factory()
        new_Person = person_factory()
        request = self.api_factory.patch("", data={"to_entry": new_Person.id})
        request.user = UserFactory(is_active=True, is_superuser=True)
        view = RelationshipModelViewSet.as_view({"patch": "partial_update"})
        # Act
        response = view(request, pk=relationship.id).render()
        # Assert
        assert response.status_code == status.HTTP_200_OK
        assert response.data["instance"]["to_entry"] == new_Person.id


# =====================================================================================================================
#                                                  TESTING CLIENT MANAGER VIEWSETS
# =====================================================================================================================


@pytest.mark.django_db
class TestClientManagerViewSet:
    @pytest.mark.parametrize("mvs", [ClientManagerViewSet])
    def test_None_qs(self, mvs):
        request = APIRequestFactory().get("")
        request.user = UserFactory(is_active=True, is_superuser=False)
        obj = ClientManagerRelationshipFactory()
        kwargs = get_kwargs(obj, mvs, request)
        vs = mvs.as_view({"get": "list"})
        response = vs(request, **kwargs)
        assert response.status_code == status.HTTP_403_FORBIDDEN

    @pytest.mark.parametrize("mvs", [ClientManagerViewSet])
    def test_delete(self, mvs):
        request = APIRequestFactory().delete("")
        request.user = UserFactory(is_active=True, is_superuser=True)
        obj1 = ClientManagerRelationshipFactory()
        obj2 = ClientManagerRelationshipFactory(client=obj1.client, status=CMR.Status.DRAFT)
        view = mvs.as_view({"delete": "destroy"})
        response = view(request, pk=obj2.id).render()
        assert response.status_code == status.HTTP_204_NO_CONTENT

    @pytest.mark.parametrize("mvs", [ClientManagerViewSet])
    def test_get(self, mvs):
        request = APIRequestFactory().get("")
        request.user = UserFactory(is_active=True, is_superuser=True)
        ClientManagerRelationshipFactory.create_batch(3)
        vs = mvs.as_view({"get": "list"})
        response = vs(request)
        assert response.data.get("results")
        assert not response.data.get("instance")
        assert len(response.data.get("results")) == 3
        assert response.status_code == status.HTTP_200_OK

    @pytest.mark.parametrize("mvs", [ClientManagerViewSet])
    def test_retrieve(self, mvs):
        request = APIRequestFactory().get("")
        request.user = UserFactory(is_active=True, is_superuser=True)
        obj = ClientManagerRelationshipFactory()
        vs = mvs.as_view({"get": "retrieve"})
        response = vs(request, pk=obj.pk)
        assert response.data.get("instance")
        assert not response.data.get("results")
        assert response.status_code == status.HTTP_200_OK

    @pytest.mark.parametrize("mvs", [ClientManagerViewSet])
    def test_put(self, mvs):
        obj_old = ClientManagerRelationshipFactory(status=CMR.Status.DRAFT)
        obj_new = ClientManagerRelationshipFactory()
        user = UserFactory(is_active=True, is_superuser=True)
        data = get_data_from_factory(obj_new, mvs, superuser=user, delete=True)
        request = APIRequestFactory().put("", data=data)
        request.user = user
        kwargs = get_kwargs(obj_old, mvs, request, data)
        vs = mvs.as_view({"put": "update"})
        response = vs(request, pk=obj_old.id, **kwargs).render()
        assert response.status_code == status.HTTP_200_OK
        assert not obj_old.client == obj_new.client
        assert response.data["instance"]["id"] == obj_old.id
        assert response.data["instance"]["client"] == obj_new.client.id
        assert not response.data["instance"]["client"] == obj_old.client.id

    @pytest.mark.parametrize("mvs", [ClientManagerViewSet])
    def test_post(self, mvs):
        cmr = ClientManagerRelationshipFactory()
        user = UserFactory(is_active=True, is_superuser=True)
        data = get_data_from_factory(cmr, mvs, delete=True, superuser=user)
        request = APIRequestFactory().post("", data=data)
        request.user = user
        kwargs = {}
        view = mvs.as_view({"post": "create"})
        response = view(request, kwargs).render()
        assert response.status_code == status.HTTP_201_CREATED
        assert response.data.get("instance")
        assert response.data["instance"]["client"] == cmr.client.id


@pytest.mark.django_db
class TestUserIsClientViewSet:
    @pytest.mark.parametrize("mvs", [UserIsClientViewSet])
    def test_get(self, mvs):
        request = APIRequestFactory().get("")
        request.user = UserFactory(is_active=True, is_superuser=True)
        obj = PersonFactory()
        user = Person.get_or_create_with_user(request.user)
        ClientManagerRelationshipFactory(client=user, relationship_manager=obj)
        vs = mvs.as_view({"get": "list"})
        response = vs(request)
        assert response.data.get("results")
        assert not response.data.get("instance")
        assert len(response.data.get("results")) == 1
        assert response.status_code == status.HTTP_200_OK

    @pytest.mark.parametrize("mvs", [UserIsClientViewSet])
    def test_retrieve(self, mvs):
        request = APIRequestFactory().get("")
        request.user = UserFactory(is_active=True, is_superuser=True)
        user = Person.get_or_create_with_user(request.user)
        rel = ClientManagerRelationshipFactory(client=user)
        vs = mvs.as_view({"get": "retrieve"})
        response = vs(request, pk=rel.pk)
        assert response.data.get("instance")
        assert not response.data.get("results")
        assert response.status_code == status.HTTP_200_OK

    @pytest.mark.parametrize("mvs", [UserIsClientViewSet])
    def test_delete(self, mvs):
        request = APIRequestFactory().delete("")
        request.user = UserFactory(is_active=True, is_superuser=True)
        obj = PersonFactory()
        user = Person.get_or_create_with_user(request.user)
        ClientManagerRelationshipFactory(client=user, relationship_manager=obj)
        view = mvs.as_view({"delete": "destroy"})
        with pytest.raises(AttributeError):
            view(request, pk=obj.id).render()


@pytest.mark.django_db
class TestUserIsManagerViewSet:
    @pytest.mark.parametrize("mvs", [UserIsManagerViewSet])
    def test_get(self, mvs):
        request = APIRequestFactory().get("")
        request.user = UserFactory(is_active=True, is_superuser=True)
        obj = EntryFactory()
        user = Person.get_or_create_with_user(request.user)
        ClientManagerRelationshipFactory(client=obj, relationship_manager=user)
        vs = mvs.as_view({"get": "list"})
        response = vs(request)
        assert response.data.get("results")
        assert not response.data.get("instance")
        assert len(response.data.get("results")) == 1
        assert response.status_code == status.HTTP_200_OK

    @pytest.mark.parametrize("mvs", [UserIsManagerViewSet])
    def test_retrieve(self, mvs):
        request = APIRequestFactory().get("")
        request.user = UserFactory(is_active=True, is_superuser=True)
        obj = EntryFactory()
        user = Person.get_or_create_with_user(request.user)
        ClientManagerRelationshipFactory(client=obj, relationship_manager=user)
        vs = mvs.as_view({"get": "retrieve"})
        response = vs(request, pk=obj.pk)
        assert response.data.get("instance")
        assert not response.data.get("results")
        assert response.status_code == status.HTTP_200_OK

    @pytest.mark.parametrize("mvs", [UserIsManagerViewSet])
    def test_delete(self, mvs):
        request = APIRequestFactory().delete("")
        request.user = UserFactory(is_active=True, is_superuser=True)
        obj = EntryFactory()
        user = Person.get_or_create_with_user(request.user)
        ClientManagerRelationshipFactory(client=obj, relationship_manager=user)
        view = mvs.as_view({"delete": "destroy"})
        response = view(request, pk=obj.id).render()
        assert response.status_code == status.HTTP_204_NO_CONTENT


@pytest.mark.django_db
class TestContactViewsets:
    @pytest.mark.parametrize(
        "mvs",
        [
            EmailContactEntryViewSet,
            AddressContactEntryViewSet,
            TelephoneContactEntryViewSet,
            WebsiteContactEntryViewSet,
            BankingContactEntryViewSet,
            SocialMediaContactEntryViewSet,
        ],
    )
    def test_get_propagated_contact(self, mvs, internal_user_factory):
        request = APIRequestFactory().get("")
        request.user = get_or_create_superuser()
        factory = get_model_factory(mvs.queryset.model)
        obj = factory(entry=internal_user_factory().profile)
        kwargs = get_kwargs(obj, mvs, request)
        vs = mvs.as_view({"get": "list"})
        response = vs(request, **kwargs)
        assert response.status_code == status.HTTP_200_OK
        assert response.data
        assert response.data.get("results")

    @pytest.mark.parametrize(
        "mvs",
        [
            EmailContactEntryViewSet,
            AddressContactEntryViewSet,
            TelephoneContactEntryViewSet,
            WebsiteContactEntryViewSet,
            BankingContactEntryViewSet,
            SocialMediaContactEntryViewSet,
        ],
    )
    def test_post_delete_contact(self, mvs):
        factory = get_model_factory(mvs.queryset.model)
        obj = factory()
        user = get_or_create_superuser()
        data = get_data_from_factory(obj, mvs, delete=True, superuser=user)
        request = APIRequestFactory().post("", data)
        request.user = user
        kwargs = get_kwargs(obj, mvs, request=request, data=data)
        vs = mvs.as_view({"post": "create"})
        response = vs(request, **kwargs)
        assert response.status_code == status.HTTP_201_CREATED
        assert response.data.get("instance")

    @pytest.mark.parametrize(
        "mvs",
        [
            EmailContactEntryViewSet,
            AddressContactEntryViewSet,
            TelephoneContactEntryViewSet,
            WebsiteContactEntryViewSet,
            BankingContactEntryViewSet,
            SocialMediaContactEntryViewSet,
        ],
    )
    def test_primary_DeleteEndpointMixin(self, mvs):
        request = APIRequestFactory().delete("")
        request.user = get_or_create_superuser()
        factory = get_model_factory(mvs.queryset.model)
        obj = factory()
        obj.primary = False
        obj.save()
        kwargs = get_kwargs(obj, mvs, request)
        vs = mvs.as_view({"delete": "destroy_multiple"})
        response = vs(request=request, **kwargs, pk=obj.pk)
        assert response.status_code == status.HTTP_204_NO_CONTENT



from unittest.mock import patch

import pytest
from wbcore.contrib.directory.factories import PersonFactory
from wbcore.contrib.workflow.dispatch import check_workflow_for_instance
from wbcore.contrib.workflow.factories import UserStepFactory
from wbcore.contrib.workflow.models import Process, ProcessStep


@pytest.mark.django_db
class TestDispatch:
    @patch("wbcore.contrib.workflow.models.step.Step.set_failed")
    def test_check_workflow_for_instance_fail_process_steps(
        self, mock_failed, process_step_factory, random_child_step_factory
    ):
        instance = PersonFactory()
        step = random_child_step_factory(exclude_factories=[UserStepFactory])
        process_step1 = process_step_factory(
            step=step,
            state=ProcessStep.StepState.WAITING,
            process__state=Process.ProcessState.ACTIVE,
            process__instance=instance,
        )
        process_step2 = process_step_factory(
            step=step,
            state=ProcessStep.StepState.ACTIVE,
            process__state=Process.ProcessState.ACTIVE,
            process__instance=instance,
        )
        process_step_factory(
            step=step,
            state=ProcessStep.StepState.ACTIVE,
            process__state=Process.ProcessState.ACTIVE,
            process__instance=instance,
            status=instance.first_name,
        )
        process_step_factory(
            step=step,
            state=ProcessStep.StepState.CANCELED,
            process__state=Process.ProcessState.ACTIVE,
            process__instance=instance,
        )
        process_step_factory(
            step=step,
            state=ProcessStep.StepState.FAILED,
            process__state=Process.ProcessState.ACTIVE,
            process__instance=instance,
        )
        process_step_factory(
            step=step,
            state=ProcessStep.StepState.FINISHED,
            process__state=Process.ProcessState.ACTIVE,
            process__instance=instance,
        )
        process_step_factory(
            step=step,
            state=ProcessStep.StepState.ACTIVE,
            process__state=Process.ProcessState.FAILED,
            process__instance=instance,
        )
        process_step_factory(
            step=step,
            state=ProcessStep.StepState.ACTIVE,
            process__state=Process.ProcessState.FINISHED,
            process__instance=instance,
        )
        process_step_factory(
            step=step,
            state=ProcessStep.StepState.ACTIVE,
            process__state=Process.ProcessState.ACTIVE,
            process__instance=PersonFactory(),
        )
        check_workflow_for_instance(instance.__class__, instance, False)
        assert len(mock_failed.call_args_list) == 2
        assert str(mock_failed.call_args_list[0][0][0].pk) == process_step1.pk
        assert str(mock_failed.call_args_list[1][0][0].pk) == process_step2.pk

    @patch("wbcore.contrib.workflow.models.workflow.Workflow.start_workflow")
    @patch("wbcore.contrib.workflow.models.workflow.Workflow.get_start_steps_for_instance")
    def test_check_workflow_for_instance_start_workflow(self, mock_start_steps, mock_start, start_step_factory):
        instance = PersonFactory()
        start_step1 = start_step_factory()
        start_step2 = start_step_factory()
        start_step_factory()
        mock_start_steps.return_value = [start_step1, start_step2]
        check_workflow_for_instance(instance.__class__, instance, False)
        assert mock_start_steps.call_args.args == (instance,)
        assert mock_start.call_args_list == [((start_step1, instance),), ((start_step2, instance),)]



from wbcore import serializers
from wbcore.contrib.notifications.models import Notification
from wbcore.contrib.notifications.serializers.notification_types import (
    NotificationTypeRepresentationSerializer,
)


class NotificationModelSerializer(serializers.ModelSerializer):
    _notification_type = NotificationTypeRepresentationSerializer(source="notification_type")

    @serializers.register_resource()
    def register_buttons(self, instance, request, user) -> dict[str, str]:
        if endpoint := instance.endpoint:
            if instance.is_endpoint_internal:
                return {"open_internal_resource": endpoint}
            else:
                return {"open_external_resource": endpoint}
        return {}

    class Meta:
        model = Notification
        fields = read_only_fields = (
            "id",
            "notification_type",
            "_notification_type",
            "title",
            "body",
            "user",
            "endpoint",
            "sent",
            "read",
            "_additional_resources",
        )



import pytest
from rest_framework.test import APIRequestFactory

from .. import resolve_profile
from ..factories import UserFactory


@pytest.mark.django_db
class TestSpecifics:
    def test_resolve_profile(self):
        request = APIRequestFactory().get("")
        request.user = UserFactory(is_active=True, is_superuser=True)
        result = resolve_profile(request)
        assert result
        assert result.get("endpoint")



from dataclasses import dataclass
from typing import Optional

from wbcore.metadata.utils import prefix_key

from .enums import ButtonDefaultColor


@dataclass
class ButtonConfig:
    label: Optional[str] = None
    icon: Optional[str] = None
    title: Optional[str] = None
    style: Optional[dict] = None
    color: ButtonDefaultColor | str = ButtonDefaultColor.PRIMARY
    weight: int = 100
    disabled: bool = False
    always_render: bool = False
    placeholder: str | None = None

    def __hash__(self):
        return hash(self.title)

    def __post_init__(self):
        if post_init := getattr(super(), "__post_init__", None):
            post_init()

        assert self.label or self.icon, "Either label or icon has to be defined."

    def __iter__(self):
        if iter := getattr(super(), "__iter__", None):
            yield from iter()

        for key in ["label", "icon", "title", "style"]:
            value = getattr(self, key, None)
            if value:
                yield key, value
        color = getattr(self.color, "value", self.color)
        yield "color", color
        yield "level", color  # TODO: we return level for backward compatibility reason. to be removed once we move to frontend version 2
        yield "disabled", self.disabled  # set to True if you want to set the css "disabled" property to that button
        yield "always_render", self.always_render  # set to True the button always needs to be rendered (even if empty)
        yield "placeholder", self.placeholder  # set to a valid string if a placeholder is needed onhover

    def serialize(self, request, key_prefix: str = None):
        res = dict(self)
        if key_prefix and "key" in res:
            res["key"] = prefix_key(res["key"], key_prefix)
        return res


@dataclass
class ButtonTypeMixin:
    def __post_init__(self):
        if post_init := getattr(super(), "__post_init__", None):
            post_init()

        assert hasattr(self, "button_type"), "button_type cannot be None."

    def __iter__(self):
        if iter := getattr(super(), "__iter__", None):
            yield from iter()

        if button_type := getattr(self, "button_type", None):
            yield "type", button_type.value


@dataclass
class ButtonUrlMixin:
    key: Optional[str] = None
    endpoint: Optional[str] = None

    def __hash__(self):
        if self.key:
            return hash(self.key)
        else:
            return hash(self.endpoint)

    def __post_init__(self):
        if post_init := getattr(super(), "__post_init__", None):
            post_init()

        assert bool(self.key) != bool(self.endpoint), "Either key or endpoint has to be defined. (Not both)"

    def __iter__(self):
        if iter := getattr(super(), "__iter__", None):
            yield from iter()
        if self.key:
            yield "key", self.key
        if self.endpoint:
            yield "endpoint", self.endpoint



from django.db.models.query import F, QuerySet
from rest_framework import status
from rest_framework.request import Request
from rest_framework.response import Response
from wbcore import viewsets
from wbcore.contrib.notifications.models import (
    NotificationType,
    NotificationTypeSetting,
)
from wbcore.contrib.notifications.serializers import (
    NotificationTypeRepresentationSerializer,
    NotificationTypeSettingModelSerializer,
)

from .configs.notification_types import NotificationTypeSettingDisplayConfig


class NotificationTypeRepresentationViewSet(viewsets.RepresentationViewSet):
    queryset = NotificationType.objects.all()
    serializer_class = NotificationTypeRepresentationSerializer


class NotificationTypeSettingModelViewSet(viewsets.ModelViewSet):
    queryset = NotificationTypeSetting.objects.all()
    serializer_class = NotificationTypeSettingModelSerializer

    display_config_class = NotificationTypeSettingDisplayConfig

    search_fields = ["notification_type__title", "notification_type__help_text"]

    def create(self, request: Request) -> Response:
        return Response(status=status.HTTP_403_FORBIDDEN)

    def destroy(self, request: Request, pk=None) -> Response:
        return Response(status=status.HTTP_403_FORBIDDEN)

    def get_queryset(self) -> QuerySet[NotificationTypeSetting]:
        return (
            super().get_queryset().filter(user=self.request.user).annotate(help_text=F("notification_type__help_text"))
        )



from django.db.models import QuerySet
from wbcore import viewsets
from wbcore.contrib.example_app.filters import StadiumFilter
from wbcore.contrib.example_app.models import Stadium
from wbcore.contrib.example_app.serializers import (
    StadiumModelSerializer,
    StadiumRepresentationSerializer,
)
from wbcore.contrib.example_app.viewsets.displays import StadiumDisplayConfig
from wbcore.contrib.example_app.viewsets.titles import StadiumTitleConfig


class StadiumModelViewSet(viewsets.ModelViewSet):
    display_config_class = StadiumDisplayConfig
    queryset = Stadium.objects.all()
    search_fields = ("name", "city__representation")
    serializer_class = StadiumModelSerializer
    title_config_class = StadiumTitleConfig
    ordering_fields = ("name", "city__representation", "capacity")
    ordering = ("name", "city__representation")
    filterset_class = StadiumFilter

    def get_queryset(self) -> QuerySet[Stadium]:
        return super().get_queryset().select_related("city")


class StadiumRepresentationViewSet(viewsets.RepresentationViewSet):
    queryset = Stadium.objects.all()
    serializer_class = StadiumRepresentationSerializer
    search_fields = ("name",)



from unittest.mock import patch

from django.dispatch import Signal
from wbcore.utils.signals import DisconnectSignals, SignalWrapper

test_signal1 = Signal()
test_signal2 = Signal()


def receiver_dummy1(sender, **kwargs):
    pass


def receiver_dummy2(sender, **kwargs):
    pass


def test_disconnect_signals():
    with patch("wbcore.tests.test_utils.test_signals.receiver_dummy1") as mock_receiver_dummy1, patch(
        "wbcore.tests.test_utils.test_signals.receiver_dummy2"
    ) as mock_receiver_dummy2:
        test_signal1.connect(receiver_dummy1, "suppressed_sender")
        test_signal1.connect(receiver_dummy1, "working_sender")
        test_signal2.connect(receiver_dummy2, "another_suppressed_sender")

        with DisconnectSignals(
            [
                SignalWrapper(test_signal1, receiver_dummy1, "suppressed_sender"),
                SignalWrapper(test_signal2, receiver_dummy2, "another_suppressed_sender"),
            ]
        ):
            test_signal1.send(sender="suppressed_sender")
            test_signal1.send(sender="working_sender")
            test_signal2.send(sender="another_suppressed_sender")

        assert mock_receiver_dummy1.call_count == 1
        assert mock_receiver_dummy1.call_args.kwargs["sender"] == "working_sender"
        assert not mock_receiver_dummy2.called



import importlib
import json
import logging
import os
import sys
import traceback
import uuid
from contextlib import suppress
from copy import deepcopy
from datetime import datetime
from importlib import import_module
from typing import Any, Callable, Dict, Optional

import magic
from celery import chain, shared_task
from croniter import croniter
from django.apps import apps
from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.core.files.base import ContentFile
from django.db import models, transaction
from django.db.models import Q
from django.db.models.signals import post_delete, pre_delete
from django.db.utils import IntegrityError
from django.dispatch import receiver
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from django_celery_beat.models import CrontabSchedule, PeriodicTask, cronexp
from import_export.resources import Resource
from picklefield.fields import PickledObjectField
from tablib import Dataset
from tqdm import tqdm
from wbcore.contrib.notifications.dispatch import send_notification
from wbcore.contrib.notifications.utils import create_notification_type
from wbcore.utils.importlib import import_from_dotted_path
from wbcore.utils.models import ComplexToStringMixin

from .enums import ExportFormat, get_django_import_export_format
from .exceptions import ImportError

logger = logging.getLogger("io")


class ParserHandler(models.Model):
    class MimeTypeChoices(models.TextChoices):
        CSV = "text/csv", "CSV"
        JSON = "application/json", "JSON"
        XLS = "application/vnd.ms-excel", "XLS"
        XLSX = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet", "XLSX"

    parser = models.CharField(max_length=255)
    handler = models.CharField(max_length=255)

    allow_file_type = models.CharField(max_length=68, blank=True, null=True, choices=MimeTypeChoices.choices)

    def __str__(self) -> str:
        return f"{self.parser}::{self.handler}"

    def parse(self, import_source: "ImportSource") -> Dict[str, Any]:
        """
        call the parser on the provided data
        Args:
            import_source: The import source containing the data
        Returns:
            The parsed data
        """

        parser = importlib.import_module(self.parser)
        return parser.parse(import_source)

    def handle(self, import_source: "ImportSource", parsed_data: Dict[str, Any], **kwargs):
        """
        Call the Handler on the parsed data
        Args:
            import_source: The initial import source
            parsed_data: The parsed_data
        """
        if parsed_data:
            model: Any = apps.get_model(self.handler)
            if handler_class := getattr(model, "import_export_handler_class", None):
                handler = handler_class(import_source)
                handler.process(parsed_data, **kwargs)

    class Meta:
        unique_together = ("parser", "handler")
        verbose_name = "Parser-Handler"
        verbose_name_plural = "Parsers-Handlers"

    @classmethod
    def get_representation_value_key(self) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(self) -> str:
        return "{{parser}}::{{handler}}"

    @classmethod
    def get_representation_endpoint(self) -> str:
        return "wbcore:io:parserhandlerrepresentation-list"


class ImportedObjectProviderRelationship(ComplexToStringMixin, models.Model):
    """
    A model that represent the relationship/link between the imported object and a provider.

    This model can be used to define different identifier in case the object is imported from difference providers
    """

    provider = models.ForeignKey(
        to="io.Provider", related_name="imported_object_relationships", on_delete=models.CASCADE
    )
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey("content_type", "object_id")
    provider_identifier = models.CharField(max_length=255)

    def compute_str(self) -> str:
        try:
            return f"{self.provider.title} -> {self.content_object}: {self.provider_identifier}"
        except AttributeError:
            return f"{self.provider.title} ({self.provider_identifier})"

    class Meta:
        verbose_name = "Content object Provider Identifier relationship"
        verbose_name_plural = "Content object Provider Identifier relationships"
        unique_together = [
            ("content_type", "object_id", "provider"),
            ("content_type", "provider_identifier", "provider"),
        ]
        indexes = [
            models.Index(fields=["content_type", "object_id", "provider"]),
            models.Index(fields=["content_type", "provider_identifier", "provider"]),
        ]


class Provider(models.Model):
    """
    Represent a data vendor/provider. Every source is linked to a provider and are represented by a unique identifier
    extracted from the data backend root folder

    Example:
        io/backend/refinitiv/instrument_prices.py => Provider.objects.get(key="refinitiv")
    """

    title = models.CharField(max_length=255)
    key = models.CharField(max_length=255, unique=True)

    def __str__(self) -> str:
        return f"{self.title} ({self.key})"

    class Meta:
        verbose_name = "Provider"
        verbose_name_plural = "Providers"


class DataBackend(models.Model):
    """
    Represents the instantiated backend imported through the specified dotted path and the passed parameters
    """

    title = models.CharField(max_length=255)

    save_data_in_import_source = models.BooleanField(
        default=True, help_text="If true, save the data in the import_source json field"
    )
    passive_only = models.BooleanField(
        default=True, help_text="If True, this data backend is allowed to be called only from the import source."
    )

    backend_class_path = models.CharField(max_length=512)
    backend_class_name = models.CharField(max_length=128, default="DataBackend")

    provider = models.ForeignKey(
        "io.Provider",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        verbose_name=_("Provider"),
    )

    def get_internal_object(self, object_content_type: ContentType, provider_external_id: str) -> models.Model | None:
        """
        Internal facing function that returns the internal id of the object with the given external provider identifier. If the relationship already exists, we simply returns the stored id. Otherwise we call the appropriate function to get
        Args:
            object_content_type: ContentType object representing the seeked object id
            provider_external_id: The object identifier as given by this backend provider

        Returns:
            The object as stored in the importer object-provider relationship. None if it doesn't exist yet
        """
        if rel := ImportedObjectProviderRelationship.objects.filter(
            provider=self.provider, provider_identifier=provider_external_id, content_type=object_content_type
        ).first():
            return rel.content_object

    class Meta:
        verbose_name = "Data Backend"
        verbose_name_plural = "Data Backends"

    @cached_property
    def backend_class(self) -> Any:
        """
        Return the imported backend class
        Returns:
            The backend class
        """
        return getattr(import_module(self.backend_class_path), self.backend_class_name)

    def __str__(self) -> str:
        return self.title

    @classmethod
    def get_representation_value_key(self) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(self) -> str:
        return "{{title}}"

    @classmethod
    def get_representation_endpoint(self) -> str:
        return "wbcore:io:databackendrepresentation-list"


class Source(models.Model):
    title = models.CharField(max_length=128)
    uuid = models.UUIDField(default=uuid.uuid4, editable=False, unique=True)
    parser_handler = models.ManyToManyField(
        "io.ParserHandler",
        verbose_name="Parser/Handler",
        related_name="sources",
    )

    is_active = models.BooleanField(default=True)

    connection_parameters = models.JSONField(default=dict, null=True, blank=True)
    import_parameters = models.JSONField(default=dict, null=True, blank=True)

    credentials = models.ManyToManyField("io.ImportCredential", blank=True, related_name="sources")

    data_backend = models.ForeignKey("io.DataBackend", on_delete=models.CASCADE, verbose_name=_("Data Backend"))

    crontab = models.ForeignKey(
        CrontabSchedule,
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        verbose_name=_("Crontab Schedule"),
        help_text=_("Crontab Schedule to run the task on.  " "Set only one schedule type, leave the others null."),
    )

    periodic_task = models.ForeignKey(
        PeriodicTask,
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        verbose_name="Periodic Task",
        related_name="sources",
    )

    import_timedelta_interval = models.IntegerField(default=0)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.instantiated_backend: Any = None

    def init_backend(self, execution_datetime: datetime):
        """
        Singleton that instantiate the backend attribute

        Args:
            execution_datetime: Datetime at which the backend is supposed to be initialize

        """
        if not self.instantiated_backend:
            self.instantiated_backend = self.data_backend.backend_class(
                import_credential=self.get_valid_credential(execution_datetime),
                data_backend=self.data_backend,
                **self.connection_parameters,
            )

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        self.update_periodic_task()

    def __str__(self) -> str:
        return self.title if self.title else f"{self.uuid}"

    class Meta:
        verbose_name = "Source"
        verbose_name_plural = "Sources"

    @property
    def current_valid_credential(self) -> "ImportCredential":
        """
        Property that returns the current valid credential

        Returns:
            The valid ImportCredential (as of now)
        """
        return self.get_valid_credential(timezone.now())

    @property
    def crontab_repr(self) -> str:
        """
        Returns the crontab string representation.
        :return: crontab string representation
        """
        if not self.crontab:
            return ""

        return "{0} {1} {2} {3} {4}".format(
            cronexp(self.crontab.minute),
            cronexp(self.crontab.hour),
            cronexp(self.crontab.day_of_month),
            cronexp(self.crontab.month_of_year),
            cronexp(self.crontab.day_of_week),
        )

    def update_periodic_task(self):
        """
        Utility function to find and update the link periodic task given the current source state
        """
        if self.crontab:
            if task := self.periodic_task:
                task.crontab = self.crontab
                task.enabled = self.is_active
                task.save()
            else:
                task = PeriodicTask.objects.update_or_create(
                    name=f"Import-Export: {self}",
                    defaults={
                        "crontab": self.crontab,
                        "args": json.dumps([self.pk]),
                        "task": "wbcore.contrib.io.models.trigger_workflow_as_task",
                    },
                )[0]
            Source.objects.filter(pk=self.pk).update(periodic_task=task)
            self.refresh_from_db()

    def get_valid_credential(self, val_datetime: datetime) -> "ImportCredential":
        """
        Get the valid credential
        Args:
            val_datetime: The date at which the credentials need to be valid
        Returns:
            A valid credential
        """
        return (
            self.credentials.filter(
                (Q(validity_start__lte=val_datetime) | Q(validity_start__isnull=True))
                & (Q(validity_end__gte=val_datetime) | Q(validity_end__isnull=True))
            )
            .order_by("-validity_end")
            .first()
        )

    def get_or_create_provider_id(self, obj: models.Model) -> str | None:
        """
        Workflow method that try to get or fetch the provider identifier for the passed object.

        If the relationship already exists, we directly returns the stored external id. Otherwise we call the appropriate function on the already instantiated backend to get it from the provider itself. If the value is returned, we store the relationship

        Args:
            obj: The object to get the external provider identifier from

        Returns:
            If it exists, the external identifier (from this source provider). None otherwise.
        """

        if (provider := self.data_backend.provider) and self.instantiated_backend is not None:
            if self.instantiated_backend.is_object_valid(obj):
                if rel := ImportedObjectProviderRelationship.objects.filter(
                    provider=provider, content_type=ContentType.objects.get_for_model(obj), object_id=obj.pk
                ).first():
                    return rel.provider_identifier
                else:
                    if provider_identifier := self.instantiated_backend.get_provider_id(obj):
                        with transaction.atomic():
                            ImportedObjectProviderRelationship.objects.create(
                                provider=provider,
                                content_type=ContentType.objects.get_for_model(obj),
                                object_id=obj.pk,
                                provider_identifier=provider_identifier,
                            )
                            return provider_identifier
            obj.save()
        else:
            raise ValueError(
                f"You can't create a provider relationship with {obj} with for backend {self.data_backend} without a provider assigned"
            )

    def is_valid_date(self, sync_datetime: datetime) -> bool:
        """
        check wether a date is valid given the stored crontab schedule
        Args:
            sync_datetime: The datetime at which validity needs to be checked
        Returns:
            True if the given date is valid given the source crontab
        """
        if not self.crontab:
            return False
        return croniter.match(self.crontab_repr, sync_datetime)

    def trigger_workflow(
        self,
        execution_time: Optional[datetime] = None,
        callback_signature: Callable | None = None,
        callback_args: list | None = None,
        callback_kwargs: Dict | None = None,
        synchronous: bool = False,
        **kwargs,
    ):
        """
        The entry point function of the whole import source workflow.
        Loop over all files return by the attached backend and save the result into a serie of import sources.

        Args:
            execution_time: The time at which the worfklow was triggerd
            callback_signature: The celery canvas signature to be called upon completion (if any)
            callback_args: the potential positional arguments to be passed to the callback function (if any)
            callback_kwargs: Potential keyword arguments to be passed to the callback function (if any)
            synchronous: If true, will execute the workflow synchronously. Default to False
            **kwargs: keyword arguments to be passed down along the workflow

        Returns:
            A generator of import sources
        """
        if not execution_time:
            execution_time = timezone.now()
        # If a queryset is provided, we need to convert it into a list of valid external id that we will get from the relationship through model or from the backend directly if not existing

        chained_tasks = [
            generate_import_sources_as_task.s(self.pk, execution_time, **kwargs),
            process_import_sources_as_task.s(),
        ]
        if not callback_args:
            callback_args = []
        if not callback_kwargs:
            callback_kwargs = {}
        if callback_signature:
            chained_tasks.append(callback_signature.si(*callback_args, **callback_kwargs))
        if synchronous:
            import_source_ids = chained_tasks[0].apply().get()
            chained_tasks[1].apply((import_source_ids,))
            if len(chained_tasks) > 2:
                chained_tasks[2].apply(tuple(callback_args), callback_kwargs)
        else:
            chain(*chained_tasks).apply_async(eta=execution_time)

    def generate_import_sources(
        self,
        execution_time: datetime,
        only_handler: str | None = None,
        **kwargs,
    ) -> list["ImportSource"]:
        """
                Process the source based on its attached backend (returns an error if such backend does not exist)
        The backend is first initialize through constructor and then, The generator `get_files` is called, passing the
        execution time, the expected data time, and some keyword arguments.

        For each file returns by the generator, we create a ImportSource objects and trigger its import task
        Args:
            execution_time: The time at which this task was triggered
            **kwargs: keyword arguments to be passed down along the workflow
        Returns:
            Return a generator of import sources
        """
        if not self.data_backend:
            raise ValueError("Data Backend not specied for this source")

        # Import the data backend module and extract its class_name
        res = []
        # Get the expected date at which this source should return its data (e.g. T-1)

        # Loop over the backend returned generator
        self.init_backend(execution_time)
        errors = []
        queryset = kwargs.pop("queryset", self.instantiated_backend.get_default_queryset())
        if self.data_backend.provider and queryset is not None:
            obj_external_ids = []
            for obj in tqdm(queryset, total=queryset.count()):
                try:
                    if external_id := self.get_or_create_provider_id(obj):
                        obj_external_ids.append(external_id)
                except IntegrityError:
                    errors.append(str(obj))
            kwargs["obj_external_ids"] = obj_external_ids
        else:
            kwargs["queryset"] = queryset
        streams = self.instantiated_backend.get_files(
            execution_time,
            **{
                **self.import_parameters,
                **kwargs,
            },
        )
        if streams:
            for file_name, _file in streams:
                content_file = ContentFile(_file.getvalue())
                content_file.name = file_name
                parser_handlers = self.parser_handler.all()
                if only_handler:
                    parser_handlers = parser_handlers.filter(handler__iexact=only_handler)
                for parser_handler in parser_handlers:
                    with transaction.atomic():
                        import_source = ImportSource.objects.create(
                            file=content_file,
                            parser_handler=parser_handler,
                            save_data=self.data_backend.save_data_in_import_source,
                            source=self,
                        )
                        res.append(import_source)
        return res

    @classmethod
    def load_sources_from_settings(self, settings: list[tuple[list[tuple[str, str]], str, dict[str, Any]]]):
        """
        Utility classmethod to parser sources from the settings.

        We assume data structure as follow

        [
            (
                [("module1.HandlerModel1", "module1.io.parsers.provider1.handler1")],
                "module1.io.backends.provider1.handler1.DataBackend",
                {
                    "credentials": [credential_dict],
                    "crontab": "10 9,17,1 * * *",
                },
            ),
            (
                [("module2.HandlerModel2", "module2.io.parsers.provider2.handler2")],
                "module2.io.backends.provider2.handler2.DataBackend",
            )
        ]
        Args:
            settings: The source settings to parse

        """
        for parser_handlers, data_backend_module, config in settings:
            with suppress(Exception):
                # This method is loaded when server boots. If this fail, it can prevent the server from starting.
                backend_class_path, _, backend_class_name = data_backend_module.rpartition(".")

                data_backend = DataBackend.objects.get(
                    backend_class_path=backend_class_path, backend_class_name=backend_class_name
                )
                crontab = None
                credentials = config.pop("credentials", [])
                if crontab_data := config.pop("crontab", None):
                    minute, hour, day_of_month, month_of_year, day_of_week = crontab_data.split(" ")
                    crontab_kwargs = {
                        "minute": minute,
                        "hour": hour,
                        "day_of_week": day_of_week,
                        "day_of_month": day_of_month,
                        "month_of_year": month_of_year,
                    }
                    crontabs = CrontabSchedule.objects.filter(**crontab_kwargs)
                    if crontabs.exists():
                        crontab = crontabs.first()
                    else:
                        crontab = CrontabSchedule.objects.create(**crontab_kwargs)

                source, source_created = Source.objects.get_or_create(
                    data_backend=data_backend,
                    defaults={"crontab": crontab, **config},
                )
                for handler, parser in parser_handlers:
                    parser_handler = ParserHandler.objects.get_or_create(parser=parser, handler=handler)[0]
                    if parser_handler not in source.parser_handler.all():
                        source.parser_handler.add(parser_handler)
                for credential_data in credentials:
                    credential, created = ImportCredential.objects.get_or_create(
                        key=credential_data["key"], defaults=credential_data
                    )
                    if credential not in source.credentials.all():
                        source.credentials.add(credential)
                source.save()

    @classmethod
    def get_representation_value_key(self) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(self) -> str:
        return "{{title}} ({{id}})"

    @classmethod
    def get_representation_endpoint(self) -> str:
        return "wbcore:io:sourcerepresentation-list"


@receiver(post_delete, sender="io.Source")
def post_delete_source(sender, instance, **kwargs):
    # Delete the attached period task
    if instance.periodic_task:
        instance.periodic_task.delete()


@receiver(pre_delete, sender="io.Source")
def pre_delete_source(sender, instance, **kwargs):
    # Unassign import source from the deleting source
    ImportSource.objects.filter(source=instance).update(source=None)


@shared_task(queue="importexport")
def trigger_workflow_as_task(source_id: int, execution_time: datetime | None = None, **kwargs):
    """
    Call the `import_source` as a celery task
    """
    source = Source.objects.get(id=source_id)
    source.trigger_workflow(execution_time=execution_time, **kwargs)


@shared_task(queue="importexport")
def generate_import_sources_as_task(source_id: int, execution_time: datetime, **kwargs) -> list[int]:
    """
    Call the `import_source` as a celery task
    """
    source = Source.objects.get(id=source_id)
    import_source_ids = [
        import_source.pk for import_source in source.generate_import_sources(execution_time, **kwargs)
    ]  # convert to list to be json serializable
    return import_source_ids


@shared_task(queue="importexport")
def process_import_sources_as_task(import_source_ids: list[int]):
    """
    Call the `import_source` as a celery task
    """
    errors = []
    for import_source_id in import_source_ids:
        import_source = ImportSource.objects.get(id=import_source_id)
        try:
            import_source.import_data(force_reimport=True)
        except ImportError:
            errors.append(import_source_id)
    if len(errors) > 0:
        raise ImportError(f"Error while processing import source {errors}")


class ImportExportSource(models.Model):
    class Status(models.TextChoices):
        PENDING = "PENDING", "Pending"
        PROCESSED = "PROCESSED", "Processed"
        WARNING = "WARNING", "Warning"
        ERROR = "ERROR", "Error"
        IGNORED = "IGNORED", "Ignore"

    created = models.DateTimeField(auto_now_add=True)
    last_updated = models.DateTimeField(auto_now=True)

    status = models.CharField(
        max_length=16,
        choices=Status.choices,
        default=Status.PENDING.value,
    )

    log = models.TextField(null=True, blank=True)
    origin = models.CharField(max_length=255, null=True, blank=True)
    creator = models.ForeignKey(
        "authentication.User", null=True, blank=True, verbose_name="Creator", on_delete=models.SET_NULL
    )
    data = models.JSONField(default=dict, null=True, blank=True)

    class Meta:
        abstract = True


class ExportSource(ImportExportSource):
    file = models.FileField(max_length=256, upload_to="io/export_source/files", null=True, blank=True)
    format = models.IntegerField(
        choices=ExportFormat.choices,
        default=ExportFormat.CSV.value,
        verbose_name=_("Format of file to be exported"),
    )

    content_type = models.ForeignKey(ContentType, verbose_name=_("Export job Content Type"), on_delete=models.CASCADE)

    resource_path = models.CharField(
        verbose_name=_("Resource path to use when exporting"), max_length=255, default="", blank=True, null=True
    )

    resource_kwargs = models.JSONField(default=dict, blank=True, null=False)

    query_str = models.TextField(verbose_name=_("SQL query to be executed"), blank=True, null=False)
    query_params = PickledObjectField(  # we have to use picklefield because the sql parameters are given as python object. However, no django model should be stored in this field, which should be the case as these objects are given through their IDs
        blank=True, verbose_name=_("SQL query parameters to be used with the sql query"), default=list
    )

    class Meta:
        verbose_name = _("Export Source")
        verbose_name_plural = _("Export Sources")
        notification_types = [
            create_notification_type(
                code="io.export_done",
                title="File Export has finished",
                help_text="Notifies user when their submitted export files is done and available",
                web=True,
                mobile=True,
                email=True,
            )
        ]
        constraints = [
            models.CheckConstraint(
                check=Q(query_str__isnull=False, resource_path__isnull=False) | ~Q(data__exact=dict()),
                name="check_either_data_or_resource_isnotnull",
            )
        ]

    @property
    def file_format(self):
        return get_django_import_export_format(self.format)()

    @property
    def resource(self) -> Resource:
        """
        Load into an attribute the instantiated resource loaded from the resource path
        """
        resource_class = import_from_dotted_path(self.resource_path)
        return resource_class(**self.resource_kwargs)

    @property
    def queryset(self) -> models.QuerySet:
        """
        Recreate the base queryset from the content type model class base manager and the saved query string and parameters
        """

        return self.content_type.model_class().objects.raw(self.query_str, self.query_params)

    def get_export_filename(self):
        date_str = timezone.now().strftime("%Y-%m-%d")
        ts = datetime.timestamp(timezone.now())
        filename = "%s-%s_%s.%s" % (
            self.content_type.model_class().__name__,
            date_str,
            ts,
            self.file_format.get_extension(),
        )
        return filename

    def notify(self):
        """
        Notify the user if the export job is done and accessible
        """
        if self.file and self.status == self.Status.PROCESSED:
            send_notification(
                code="io.export_done",
                title=_("Your export file is available"),
                body=_("<p>The export job you requested is finished and available for one hour.</p>").format(
                    url=self.file.url
                ),
                endpoint=self.file.url,
                user=self.creator,
            )

    def export_data(self, **kwargs):
        self.log = ""
        self.status = self.Status.PENDING.name
        self.save()
        try:
            file_format = get_django_import_export_format(self.format)()
            if self.data:
                # Data was saved as a dictionary from a tablib.dataset, we need to recreate it
                data = Dataset()
                data.headers = self.data["headers"]
                data.extend(self.data["data"])
            else:
                data = self.resource.export(queryset=self.queryset)

            export_data = file_format.export_data(data)
            if not file_format.is_binary():
                export_data = export_data.encode("utf-8")
            self.file.save(self.get_export_filename(), ContentFile(export_data))
            self.status = self.Status.PROCESSED.name
            self.save()
            self.notify()

        except Exception:
            ex_type, ex_value, _ = sys.exc_info()
            self.status = self.Status.ERROR.name
            self.log = f"{ex_type}: {ex_value}\n"
            self.log += traceback.format_exc()
            self.save()


class ImportSource(ImportExportSource):
    file = models.FileField(max_length=256, upload_to="io/import_source/files", null=True, blank=True)

    save_data = models.BooleanField(default=True, help_text="If True, will save the raw data in a field")
    progress_index = models.PositiveIntegerField(default=0)

    parser_handler = models.ForeignKey("io.ParserHandler", on_delete=models.CASCADE)
    source = models.ForeignKey("io.Source", null=True, blank=True, verbose_name="Source", on_delete=models.SET_NULL)
    errors_log = models.TextField(null=True, blank=True)

    class Meta:
        verbose_name = "Import Source"
        verbose_name_plural = "Import Sources"
        notification_types = [
            create_notification_type(
                code="io.import_done",
                title="File Import has finished",
                help_text="Notifies user when their submitted import files has finished",
            )
        ]

    def __str__(self) -> str:
        return f"{self.pk}"

    def _validate_file_type(self):
        """
        Valid the file type based on the allowed types specified by the parser
        Returns:
            True if the filetype is valid
        """
        if (upload := self.file) and (allow_file_type := self.parser_handler.allow_file_type):
            file_type = magic.from_buffer(upload.file.read(1024), mime=True)
            if file_type != allow_file_type:
                raise ValidationError("File type not supported.")

    def save(self, *args, **kwargs):
        # Check here company name, modify ...
        self._validate_file_type()
        super().save(*args, **kwargs)

    def _parse_data(self) -> dict[str, Any]:
        """
        Given the ParserHandler linked object, we import its related module and call its defined `parse` function
        Returns:
            Dict The parsed data as a python dictionary
        """
        if not self.file:
            raise ValueError("This import source does not include a valid file")
        parsed_data = self.parser_handler.parse(self)
        if self.save_data:
            self.data = parsed_data
        self.save()
        return parsed_data

    def _process_data(self, parsed_data: dict[str, Any], **kwargs):
        """
        Given the parsed data (as a python dictionary), import the corresponding handler and calls its defined `process
        Args:
            data: The deserialized data to be passed down to the parser
        """
        self.status = self.Status.PROCESSED.name
        if parsed_data:
            parsed_data_copy = deepcopy(parsed_data)
            if data := parsed_data_copy.pop("data", None):
                self.parser_handler.handle(self, {"data": data[self.progress_index :], **parsed_data_copy}, **kwargs)
                if self.errors_log:
                    self.status = self.Status.WARNING.name
        self.save()

    def import_data(self, force_reimport: Optional[bool] = False, **kwargs):
        """
        General workflow method:
        * Parse the data given the linked import file
        * Handle the parsed data
        * Change the import_source status based on sucess
        Args:
            silent: False if this method is not suppose to silently catch the exceptions
        """
        if force_reimport:
            self.progress_index = 0
        self.log = ""
        self.errors_log = ""
        # self.data = {}
        self.status = self.Status.PENDING.name
        self.save()
        debug = kwargs.get("debug", settings.DEBUG)
        try:
            data = self._parse_data()
            self._process_data(data, debug=debug)
        except Exception as e:
            ex_type, ex_value, _ = sys.exc_info()
            self.status = self.Status.ERROR.name
            self.log = f"{ex_type}: {ex_value}\n"
            self.log += traceback.format_exc()
            self.save()
            if not debug:
                raise ImportError(
                    f"Could not import file {self.file.name} for backend {self.source.data_backend.title} and parser/handler {self.parser_handler}"
                )
            else:
                raise e

    @classmethod
    def get_representation_value_key(self) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(self) -> str:
        return "{{file}}"

    @classmethod
    def get_representation_endpoint(self) -> str:
        return "wbcore:io:importsourcerepresentation-list"


@shared_task(queue="importexport")
def import_data_as_task(import_source_id: int, **kwargs):
    """
    Call `import_data` as a celery task
    :param Int The import_source id
    """
    import_source = ImportSource.objects.get(id=import_source_id)
    import_source.import_data(**kwargs)


@shared_task(queue="importexport")
def export_data_as_task(export_source_id: int, **kwargs):
    """
    Call `import_data` as a celery task
    :param Int The import_source id
    """
    export_source = ExportSource.objects.get(id=export_source_id)
    export_source.export_data(**kwargs)


@receiver(models.signals.post_save, sender=ExportSource)
def post_save_export_source(sender, instance: ExportSource, created: bool, raw: bool, **kwargs):
    """Triggers the export task on creation"""

    if not raw and created and instance.status == ExportSource.Status.PENDING:
        transaction.on_commit(lambda: export_data_as_task.delay(instance.id))


def validate_key_file(value: models.FileField):
    ext = os.path.splitext(value.name)[1]
    if ext.lower() != ".key":
        raise ValidationError("The file extension needs to be a .key")


def validate_pem_file(value: models.FileField):
    ext = os.path.splitext(value.name)[1]
    if ext.lower() != ".pem":
        raise ValidationError("The file extension needs to be a .pem")


class ImportCredential(models.Model):
    class Type(models.TextChoices):
        CREDENTIAL = "CREDENTIAL", "Credential"
        AUTHENTICATION_TOKEN = "AUTHENTICATION_TOKEN", "Authentication Token"
        CERTIFICATE = "CERTIFICATE", "Certificate"

    key = models.CharField(max_length=255)
    type = models.CharField(choices=Type.choices, default=Type.CREDENTIAL, max_length=64)

    username = models.CharField(max_length=255, null=True, blank=True)
    password = models.CharField(max_length=255, null=True, blank=True)

    authentication_token = models.CharField(max_length=2048, null=True, blank=True)

    certificate_pem = models.FileField(
        max_length=256,
        null=True,
        blank=True,
        upload_to="io/import_credential/certificates",
        help_text="We are expecting a .cert file",
        validators=[validate_pem_file],
    )
    certificate_key = models.FileField(
        max_length=256,
        null=True,
        blank=True,
        upload_to="io/import_credential/certificates",
        help_text="We are expecting a .key file",
        validators=[validate_key_file],
    )

    additional_resources = models.JSONField(default=dict, null=True, blank=True)

    validity_start = models.DateTimeField(null=True, blank=True)
    validity_end = models.DateTimeField(null=True, blank=True)

    def save(self, *args, **kwargs):
        if self.type == self.Type.CREDENTIAL and (not self.username or not self.password):
            raise ValidationError("The type is credential, you need to specify a username and a password")
        elif self.type == self.Type.AUTHENTICATION_TOKEN and not self.authentication_token:
            raise ValidationError("The type is authentication header, you need to specify a valid header")
        elif self.type == self.Type.CERTIFICATE and (not self.certificate_key or not self.certificate_pem):
            raise ValidationError("The type is certificate, you need to specify valid public and private key files")
        super().save(*args, **kwargs)

    class Meta:
        verbose_name = "Import Credential"
        verbose_name_plural = "Import Credentials"

    def __str__(self) -> str:
        dates_repr = ""
        if self.validity_start:
            dates_repr += f"{self.validity_start:%Y-%m-%d %H:%M:%S}"
        if self.validity_end:
            dates_repr += f"- {self.validity_end:%Y-%m-%d %H:%M:%S}"
        if dates_repr:
            return f"{self.key} ({dates_repr})"
        return self.key



from contextlib import suppress
from functools import partial

from django.contrib.postgres.fields import (
    ArrayField,
    DateRangeField,
    DateTimeRangeField,
    DecimalRangeField,
)
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.db.utils import ProgrammingError
from django.urls import resolve
from django.utils.http import urlencode
from django.utils.text import capfirst
from django_fsm import FSMField, get_available_user_FIELD_transitions
from rest_framework import serializers
from rest_framework.relations import ManyRelatedField, RelatedField
from rest_framework.reverse import NoReverseMatch, reverse
from rest_framework.serializers import PKOnlyObject, SerializerMetaclass
from timezone_field import TimeZoneField
from wbcore.contrib.color.fields import ColorField
from wbcore.contrib.icons.models import IconField
from wbcore.contrib.icons.serializers import IconSelectField
from wbcore.models.fields import YearField
from wbcore.serializers import (
    fields,
    register_only_instance_resource,
    register_resource,
)
from wbcore.serializers.fields.mixins import WBCoreSerializerFieldMixin
from wbcore.serializers.fields.related import ListSerializer
from wbcore.serializers.fields.types import WBCoreType
from wbcore.serializers.mixins import UserProfileMixin

from rest_framework.relations import Hyperlink  # NOQA # isort:skip


def _serialized_primarykey(instance, value):
    return value.id


def validate_nested_representation(instance, value):
    if isinstance(value, list):
        return [getattr(v, "id", None) for v in value]
    return getattr(value, "id", None)


class WBCoreSerializerMetaClass(SerializerMetaclass):
    def __new__(cls, name, bases, dct):
        _class = super().__new__(cls, name, bases, dct)

        if _meta := getattr(_class, "Meta", None):
            model = _meta.model
            only_fsm_transition_on_instance = getattr(_meta, "only_fsm_transition_on_instance", False)
            with suppress(AttributeError):
                for field in filter(lambda f: isinstance(f, FSMField), model._meta.fields):
                    transitions = getattr(model, f"get_all_{field.name}_transitions")(model())
                    for transition in transitions:

                        def method(self, instance, request, user, field, transition, view=None):
                            # if self.context["view"].historical_mode:
                            #     return {}

                            transitions = get_available_user_FIELD_transitions(instance, user, field)
                            if transition.name in [t.name for t in transitions]:
                                url = resolve(request.path_info)
                                namespace = f"{url.namespace}:" if url.namespace else ""
                                base_url_name = url.url_name.split("-")[:-1]

                                # We need to pass the kwargs from the view through to the reverse call
                                # And additionally pass in the instance.id as the pk
                                # NOTE: What happens if the reverse parameter is not called pk?
                                # NOTE: Is that even possible?

                                # If the view is not in the context, we just create an empty keyword dict.
                                # FIXME: This should actually never happen
                                try:
                                    kwargs = self.context["view"].kwargs
                                except KeyError:
                                    kwargs = {}

                                kwargs.update({"pk": instance.id})
                                try:
                                    endpoint = reverse(
                                        f"{namespace}{'-'.join(base_url_name)}-{transition.name}",
                                        kwargs=kwargs,
                                        request=request,
                                    )
                                except NoReverseMatch:
                                    return {}

                                return {transition.name: endpoint}
                            return {}

                        if only_fsm_transition_on_instance:
                            wrapped_method = register_only_instance_resource()(
                                partial(method, field=field, transition=transition)
                            )
                            wrapped_method.__name__ = transition.name
                        else:
                            wrapped_method = register_resource()(partial(method, field=field, transition=transition))
                            wrapped_method.__name__ = transition.name

                        setattr(
                            _class,
                            transition.name,
                            wrapped_method,
                        )

            # Flatten field from declared nested serializer
            for parent_field_name, child_fields in getattr(_meta, "flatten_fields", dict()).items():
                nested_serializer_class = getattr(child_fields, "serializer_class", None)
                nested_serializer = None
                # If a nested serializer class is provided, we get the list of flatten fields from it (e.g. JSONTableField)
                if nested_serializer_class:
                    nested_serializer = nested_serializer_class(json_serializer=True)
                    if flatten_field_names := getattr(child_fields, "flatten_field_names", []):
                        child_fields = nested_serializer.extract_representation_fields(
                            parent_field_name, flatten_field_names
                        )
                for child_field_name, serializer_field in child_fields:
                    # For each nested field, we include them in the class declaration
                    nested_field_name = f"{parent_field_name}__{child_field_name}"
                    _class._declared_fields[nested_field_name] = serializer_field
                    if nested_field_name not in _meta.fields:
                        _meta.fields = (*_meta.fields, nested_field_name)

                    # If the nested field is a related field, we check if a serializer is provided to get the RepresentationModelSerializer to attach
                    if issubclass(serializer_field.__class__, ManyRelatedField) or issubclass(
                        serializer_field.__class__, RelatedField
                    ):
                        setattr(
                            _class,
                            f"validate_{nested_field_name}",
                            lambda instance, value: validate_nested_representation(instance, value),
                        )
                        if nested_serializer and (
                            representation_serializer_field := nested_serializer.fields.get(
                                f"_{child_field_name}", None
                            )
                        ):
                            representation_serializer_field.set_attr("source", nested_field_name)
                            related_nested_field_name = f"_{nested_field_name}"
                            _class._declared_fields[related_nested_field_name] = representation_serializer_field
                            if related_nested_field_name not in _meta.fields:
                                _meta.fields = (*_meta.fields, related_nested_field_name)

        return _class


class AdditionalMetadataMixin:
    def __init__(self, *args, **kwargs):
        self.json_serializer = kwargs.pop("json_serializer", False)
        super().__init__(*args, **kwargs)

    def validate(self, data):
        data = super().validate(data)
        if hasattr(self, "Meta"):
            for parent_field in getattr(self.Meta, "flatten_fields", {}).keys():
                primary_data = getattr(self.instance, parent_field, {})
                updated_data = data.get(parent_field, {})
                primary_data.update(updated_data)
                data[parent_field] = primary_data
        return data

    def set_attr(self, attr, value):
        setattr(self, attr, value)
        self._kwargs[attr] = value

    def build_standard_field(self, field_name, model_field):
        field_class, field_kwargs = super().build_standard_field(field_name, model_field)
        if isinstance(model_field, FSMField):
            field_class = self.serializer_fsm_field
        if model_field and not field_kwargs.get("label", None) and model_field.verbose_name:
            field_kwargs["label"] = capfirst(model_field.verbose_name)
        return field_class, field_kwargs

    def build_relational_field(self, field_name, relation_info):
        model_field, related_model, to_many, to_field, has_through_model, reverse = relation_info
        field_class, field_kwargs = super().build_relational_field(field_name, relation_info)
        if has_through_model:
            field_kwargs["queryset"] = related_model.objects.all()
            field_kwargs["read_only"] = False
        if (
            relation_info.model_field
            and not field_kwargs.get("label", None)
            and relation_info.model_field.verbose_name
        ):
            field_kwargs["label"] = capfirst(model_field.verbose_name)
        return field_class, field_kwargs

    def build_property_field(self, field_name, model_class):
        field_class = fields.ReadOnlyField
        field_kwargs = {}

        return field_class, field_kwargs

    @classmethod
    def get_decorators(cls):
        if hasattr(cls, "Meta"):
            yield from getattr(cls.Meta, "decorators", dict()).items()

    @classmethod
    def get_percent_fields(cls):
        if hasattr(cls, "Meta"):
            yield from getattr(cls.Meta, "percent_fields", list())

    @classmethod
    def get_dependency_map(cls):
        if hasattr(cls, "Meta"):
            yield from getattr(cls.Meta, "dependency_map", dict()).items()

    def get_fields(self):
        fields = super().get_fields()

        # This shouldn't impact performance much has get_fields is used behind a cached property

        # Get decorators specified in Meta and append the argument accordingly
        for key, value in self.get_decorators():
            if field := fields.get(key, None):
                field.set_attr("decorators", [value])

        # Get percent fields specified in Meta and append the argument accordingly
        for key in self.get_percent_fields():
            if field := fields.get(key, None):
                field.set_percent(True)

        # Get dependency map specified in Meta and append the argument accordingly
        for key, values in self.get_dependency_map():
            for value in values:
                if depends_on_field := fields.get(key, None):
                    depends_on_field.append_depend_on({"field": value, "options": {}})

        return fields

    def extract_representation_fields(self, parent_field_name, nested_field_names):
        """
        Use to extract fields that are define on the serializer as nested fields
        """
        for flatten_field_name in nested_field_names:
            with suppress(AttributeError, ImproperlyConfigured):
                nested_serializer_field = self.fields[flatten_field_name]
                if not hasattr(nested_serializer_field, "endpoint"):
                    # It's not representation serializer field
                    nested_serializer_field.set_attr("source", f"{parent_field_name}.{flatten_field_name}")
                    nested_serializer_field.set_attr("allow_null", True)
                    nested_serializer_field.set_attr("read_only", False)
                    yield flatten_field_name, nested_serializer_field


class Serializer(AdditionalMetadataMixin, UserProfileMixin, serializers.Serializer):
    _additional_resources = fields.AdditionalResourcesField()
    _buttons = fields.DynamicButtonField()


class ModelSerializer(
    AdditionalMetadataMixin,
    UserProfileMixin,
    serializers.ModelSerializer,
    metaclass=WBCoreSerializerMetaClass,
):
    serializer_field_mapping = {
        models.AutoField: fields.PrimaryKeyField,
        models.BigAutoField: fields.PrimaryKeyField,
        models.BooleanField: fields.BooleanField,
        models.CharField: fields.CharField,
        models.DateField: fields.DateField,
        models.DateTimeField: fields.DateTimeField,
        models.TimeField: fields.TimeField,
        models.DecimalField: fields.DecimalField,
        models.DurationField: fields.DurationField,
        models.FileField: fields.FileField,
        models.FloatField: fields.FloatField,
        models.ImageField: fields.ImageField,
        models.IntegerField: fields.IntegerField,
        models.PositiveIntegerField: fields.IntegerField,
        models.PositiveSmallIntegerField: fields.IntegerField,
        models.SmallIntegerField: fields.IntegerField,
        models.TextField: fields.TextField,
        models.URLField: fields.URLField,
        models.UUIDField: fields.CharField,
        models.GenericIPAddressField: fields.CharField,
        ArrayField: fields.ListField,
        DateTimeRangeField: fields.DateTimeRangeField,
        DateRangeField: fields.DateRangeField,
        models.JSONField: fields.JSONTableField,
        fields.StarRatingField: fields.StarRatingField,
        fields.EmojiRatingField: fields.EmojiRatingField,
        ColorField: fields.ColorPickerField,
        DecimalRangeField: fields.DecimalRangeField,
        IconField: IconSelectField,
        TimeZoneField: fields.TimeZoneField,
        YearField: fields.YearField,
    }
    serializer_related_field = fields.PrimaryKeyRelatedField
    serializer_choice_field = fields.ChoiceField
    serializer_multiple_choice_field = fields.MultipleChoiceField
    serializer_fsm_field = fields.FSMStatusField

    _additional_resources = fields.AdditionalResourcesField()
    _buttons = fields.DynamicButtonField()


class RepresentationSerializer(WBCoreSerializerFieldMixin, ModelSerializer):
    field_type = WBCoreType.SELECT.value

    def __init__(self, *args, **kwargs):
        self.ignore_filter = kwargs.pop("ignore_filter", getattr(self, "ignore_filter", None))
        self.filter_params = kwargs.pop("filter_params", getattr(self, "filter_params", None))
        self.endpoint = kwargs.pop(
            "endpoint",
            getattr(self, "endpoint", getattr(self.Meta.model, "get_representation_endpoint", None)),
        )
        if callable(self.endpoint):
            self.endpoint = self.endpoint()
        self.value_key = kwargs.pop(
            "value_key",
            getattr(self, "value_key", getattr(self.Meta.model, "get_representation_value_key", None)),
        )
        if callable(self.value_key):
            self.value_key = self.value_key()
        self.label_key = kwargs.pop(
            "label_key",
            getattr(self, "label_key", getattr(self.Meta.model, "get_representation_label_key", None)),
        )
        if callable(self.label_key):
            self.label_key = self.label_key()
        self.label_key = kwargs.pop("label_key", self.label_key)
        self.optional_get_parameters = kwargs.pop(
            "optional_get_parameters",
            getattr(self, "optional_get_parameters", None),
        )
        super().__init__(*args, **kwargs)

    def to_representation(self, value):
        # In case we annotate the representation, we need to ensure that the value is an Pk object
        if isinstance(value, int):
            value = PKOnlyObject(value)
        with suppress(ProgrammingError):
            if isinstance(value, PKOnlyObject) and self.Meta.model:
                try:
                    value = self.Meta.model._default_manager.get(id=value.pk)
                except self.Meta.model.DoesNotExist:
                    value = None
        return super().to_representation(value)

    @classmethod
    def many_init(cls, *args, **kwargs):
        kwargs["child"] = cls(
            *args,
            # depends_on=kwargs.get("depends_on", None),
            **kwargs,
        )
        fields = [
            "filter_params",
            "endpoint",
            "value_key",
            "label_key",
            "optional_get_parameters",
            "extra",
            "decorators",
            "depends_on",
            "math",
            "ignore_filter",
        ]
        for field in fields:
            if field in kwargs:
                del kwargs[field]
        return ListSerializer(*args, **kwargs)

    def __new__(cls, *args, **kwargs):
        if "read_only" not in kwargs:
            kwargs["read_only"] = True
        return super().__new__(cls, *args, **kwargs)

    def _get_filter_params(self, request):
        if not self.ignore_filter:
            if self.filter_params is not None:
                if callable(self.filter_params):
                    return self.filter_params(request)
                else:
                    return self.filter_params
            elif (get_filter_parm := getattr(self, "get_filter_params", None)) and callable(get_filter_parm):
                return get_filter_parm(request)
        return {}

    def get_representation(self, request, field_name):
        _, super_representation = super().get_representation(request, field_name)
        url = reverse(self.endpoint, request=request)

        filter_params = self._get_filter_params(request)
        if isinstance(filter_params, dict):
            # Sanitize None values from the filter dictionary (otherwise urlencode will throw an error)
            filter_params = {k: v for k, v in filter_params.items() if v is not None}
            url = f"{url}?{urlencode(filter_params)}"

        representation = {
            "type": self.field_type,
            "depends_on": super_representation.get("depends_on"),
            "representation_key": field_name,
            "related_key": self.get_related_key(),
            "endpoint": {
                "url": url,
                "value_key": self.value_key,
                "label_key": self.label_key,
            },
        }

        if self.help_text:
            representation["help_text"] = self.help_text

        if self.optional_get_parameters:
            representation["endpoint"]["optional_get_parameters"] = self.optional_get_parameters
        return self.get_related_key(), representation



import factory
from django.contrib.contenttypes.models import ContentType
from django.db.models import signals
from wbcore.contrib.directory.models import Person
from wbcore.contrib.workflow.models import Workflow


@factory.django.mute_signals(signals.post_save)
class WorkflowFactory(factory.django.DjangoModelFactory):
    name = factory.Faker("word")
    status_field = "first_name"
    single_instance_execution = factory.Faker("pybool")
    preserve_instance = factory.Faker("pybool")
    model = factory.LazyAttribute(lambda o: ContentType.objects.get_for_model(Person))
    graph = factory.django.ImageField()

    class Meta:
        model = Workflow
        django_get_or_create = ["name"]



from django.utils.translation import gettext as _
from rest_framework.reverse import reverse
from wbcore.contrib.icons import WBIcon
from wbcore.enums import RequestType
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig


class ReleaseNotesButtonConfig(ButtonViewConfig):
    def get_custom_buttons(self):
        if not self.view.kwargs.get("pk", None):
            return {
                bt.ActionButton(
                    method=RequestType.PATCH,
                    action_label=_("Reading all release notes"),
                    endpoint=reverse("wbcore:releasenote-mark-all-as-read", request=self.request),
                    description_fields=_("Do you want to mark all release notes as read?"),
                    label=_("Mark all as read"),
                    icon=WBIcon.VIEW.icon,
                    confirm_config=bt.ButtonConfig(label=_("Read all")),
                    cancel_config=bt.ButtonConfig(label=_("Cancel")),
                    identifiers=("wbcore:releasenote",),
                ),
            }
        return dict()
